
import psycopg2.errors
from unittest.mock import patch
from odoo import Command
from odoo.tests import common, tagged
from odoo.tools import cloc, mute_logger
from odoo.exceptions import AccessError, ValidationError
from odoo.addons.loempia import util
from odoo.addons.loempia.util import ManifestError
from odoo.addons.loempia.tests.common import MockRequest, TestLoempiaData

import shutil
import os
from datetime import date
from io import BytesIO
from zipfile import ZipFile

@tagged('post_install', '-at_install')
class TestLoempiaModule(TestLoempiaData, common.TransactionCase):

    @classmethod
    def setUpClass(self):
        super(TestLoempiaModule, self).setUpClass()
        self.repositories = self._create_repos()

    @classmethod
    def tearDownClass(self):
        for repo in self.repositories:
            if os.path.exists(repo.path):
                shutil.rmtree(repo.path)
        super(TestLoempiaModule, self).tearDownClass()

    @classmethod
    def _create_repos(self):
        series = {"loempia.series_13": 13.0, "loempia.series_12": 12.0, "loempia.series_11": 11.0}
        Repositories = self.env['loempia.repo']
        for k, v in series.items():
            Repositories |= self.env['loempia.repo'].create({
                'series_id': self.env.ref(k).id,
                'url': 'ssh://git@github.com/ndeodoo/essai.git#%s' % str(v),
                'vcs': 'git',
                'maintainer_id': self.maintainer.id,
                'state': 'active',
            })
        return Repositories

    def _create_module(self, module_name, repo_id, **params):
        template = params.get('template_id') or self.env['loempia.module.template'].search([('name', '=', module_name), ('maintainer_id', '=', self.maintainer.id)])
        return self.env['loempia.module'].create({
            'name': module_name,
            'version': '1.0',
            'license': 'LGPL-3',
            'currency_id': self.env.ref('base.EUR').id,
            'price_single': 10.0,
            'repo_id': repo_id.id,
            'status': 'approved',
            'template_id': template.id if template else self._create_template(module_name).id,
            'depend_ids': [] if not params.get('depend_ids') else params.get('depend_ids')
        })

    def _create_template(self, module_name):
        return self.env['loempia.module.template'].create(
            {
                'name': module_name,
                'maintainer_id': self.maintainer.id,
            })

    def _create_confirmed_purchase(self, module):
        return self.env['loempia.module.purchase'].create({
            'order_id': 1,
            'order_name': 'SO0001',
            'module_id': module.id,
            'template_id': module.template_id.id,
            'product_id': module.product_id,
            'date_order': date.today(),
            'quantity': 1,
            'price_unit': module.price_single,
            'user_id': self.customer.id,
            'state': 'confirmed',
        })

    def test_01_dependencies(self):
        """
        This test ensures that the module are correclty updated if a dependency is incorrect.
        In this case the "Z" module does not exist and "D" should not depends of a non-existing
        module.

        Moreover, it also ensures that the other correct dependencies are correctly updated.
        """
        repo_id = self.repositories[0]
        modules = self.env['loempia.module']
        for name in ['A', 'B', 'C', 'D']:
            modules |= self._create_module(name, repo_id)

        values = {
            'A': {'name': 'A', 'depends': ['B']},
            'B': {'name': 'B', 'depends': ['C', 'D']},
            'C': {'name': 'C', 'depends': []},
            'D': {'name': 'D', 'depends': ['Z']}
        }

        correct_values = {
            'A': set(['B']),
            'B': set(['C', 'D']),
            'C': set([]),
            'D': set([])  # Z does not exist, so the module should not be updated this way
        }

        modules_to_update = modules.mapped('name')
        repush = None
        while modules_to_update:
            mod = modules_to_update.pop(0)
            value = values[mod]
            try:
                to_add = []
                dpds = self.env['loempia.module']._get_dependencies(repo_id, value, modules_to_update)
                to_add = [(4, dep, 0) for dep in dpds]
                modules.filtered(lambda m: m.name == mod).write({'depend_ids': to_add})
                repush = None
            except util.DependencyError:
                if repush == mod:
                    # ignore module
                    repush = None
                    continue
                if repush is None:
                    repush = mod
                modules_to_update.append(mod)

        # Ensure that depend_ids is correctly set
        for module in modules:
            correct_value = correct_values.get(module.name)
            self.assertEqual(set(module.depend_ids.mapped('name')), correct_value)

        # Ensure that all_depend_ids is correctly computed.
        module_A = modules.filtered(lambda m: m.name == 'A')
        self.assertEqual(set(module_A.all_depend_ids.mapped('name')), set(['B', 'C', 'D']))

    def test_02_dependencies(self):
        """
        This test ensure that a module with incorrect dependencies is not created if
        it does not exist previously. However, the module which are correcty parametrized
        must be created with the correct values.
        """
        repo_id = self.repositories[0]
        modules = self.env['loempia.module']

        values = {
            'A': {'name': 'A', 'depends': ['B', 'Z']},
            'B': {'name': 'B', 'depends': ['C', 'D']},
            'C': {'name': 'C', 'depends': []},
            'D': {'name': 'D', 'depends': []}
        }

        # Z does not exist so A should not be created
        correct_values = {
            'B': set(['C', 'D']),
            'C': set([]),
            'D': set([])
        }

        modules_to_update = ['A', 'B', 'C', 'D']
        repush = None
        while modules_to_update:
            mod = modules_to_update.pop(0)
            value = values[mod]
            try:
                to_add = []
                dpds = self.env['loempia.module']._get_dependencies(repo_id, value, modules_to_update)
                to_add = [(4, dep, 0) for dep in dpds]
                modules |= self._create_module(mod, repo_id, **{'depend_ids': to_add})
                repush = None
            except util.DependencyError:
                if repush == mod:
                    # ignore module
                    repush = None
                    continue
                if repush is None:
                    repush = mod
                modules_to_update.append(mod)

        # A depends of a non-existing module so it should not be created
        self.assertEqual(len(modules), 3, msg="Only modules with correct dependencies should be created")
        self.assertEqual(modules.sorted('name').mapped('name'), list(correct_values.keys()))
        for module in modules:
            correct_value = correct_values.get(module.name)
            self.assertEqual(set(module.depend_ids.mapped('name')), correct_value)

    def test_toggle_block_module(self):
        """
        This test ensures that the toggle block (module) update the status of a module correctly
        """
        repo_id = self.repositories[0]
        module = self._create_module('A', repo_id)

        # 01
        module.toggle_block()
        self.assertEqual(module.status, 'blocked', msg="If an approved module is blocked, its state has to been set to 'blocked'")

        # 02 toggle block with a banned maintainer
        module.repo_maintainer_id.banned = True
        module.toggle_block()
        self.assertEqual(module.status, 'banned', msg="Unblock a module of a banned maintainer should set its state to 'banned'")

        # 03 toggle block regular maintainer
        module.repo_maintainer_id.banned = False
        module.status = 'blocked'
        module.toggle_block()
        self.assertEqual(module.status, 'invalid', msg="Unblock a module should set its state to 'invalid'")

        # 04 Ensure portal/customer can't toggle_block
        try:
            module.with_user(self.customer.id).toggle_block()
        except AccessError:
            self.assertEqual(module.status, 'invalid', msg='portal has no access to block/unblock module')

    def test_toggle_block_template(self):
        """
        This test ensures that the toggle block (template) update the status of a module correctly
        """
        modules = self.env['loempia.module']
        for repo in self.repositories:
            modules |= self._create_module('A', repo)

        template = modules.mapped('template_id')
        self.assertEqual(len(template), 1)

        # 01 block template --> block module
        template.toggle_block()
        self.assertTrue(template.is_blocked)
        self.assertEqual(set(modules.mapped('status')), set(['blocked']))

        # 02 unblock template --> invalid module
        template.toggle_block()
        self.assertTrue(not template.is_blocked)
        self.assertEqual(set(modules.mapped('status')), set(['invalid']))

        # 03 Portal can't toggle block template
        try:
            template.with_user(self.customer.id).toggle_block()
        except AccessError:
            self.assertTrue(not template.is_blocked, msg='portal has no access to block/unblock template')

    def test_bundle_price(self):
        """
        This test ensure that the price is correctly computed
        """
        repo_id = self.repositories[0]
        module_D = self._create_module('D', repo_id)
        module_C = self._create_module('C', repo_id)

        module_B_dependencies = [(4, dep, 0) for dep in set([module_C.id, module_D.id])]
        module_B = self._create_module('B', repo_id, **{'depend_ids': module_B_dependencies})

        module_A_dependencies = [(4, dep, 0) for dep in set([module_B.id])]
        module_A = self._create_module('A', repo_id, **{'depend_ids': module_A_dependencies})

        self.assertEqual(module_D.price_single, 10.0)
        self.assertEqual(module_D.price, 10.0)  # No dpd

        self.assertEqual(module_C.price_single, 10.0)
        self.assertEqual(module_C.price, 10.0)  # No dpd

        self.assertEqual(module_B.price_single, 10.0)
        self.assertEqual(module_B.price, 30.0)  # dpd of C and D whose have 10 as price_single

        self.assertEqual(module_A.price_single, 10.0)
        self.assertEqual(module_A.price, 40.0)  # dpd of D,C and B whose have 10 as singe_price

    def test_toggle_publish(self):
        """
        This test ensure that the toggle publish correctly works
        """
        repo_id = self.repositories[0]
        module = self._create_module('A', repo_id)

        self.assertTrue(module.is_published, msg='Published default')
        module._toggle_publish()
        self.assertTrue(not module.is_published, msg='Unpublished by toggle')

    def test_is_iap(self):
        """
        This test ensure that if a module depends of iap directly or indirectly,
        then is_iap is set to True ('iap' in all_depend_ids).
        """
        repo_id = self.repositories[0]
        module_iap = self._create_module('iap', repo_id)
        module_B = self._create_module('B', repo_id)

        self.assertTrue(not module_iap.is_iap, msg="does not depend of iap")
        self.assertTrue(not module_B.is_iap, msg="does not depend of iap")
        module_A_dependencies = [(4, dep, 0) for dep in set([module_B.id, module_iap.id])]
        module_A = self._create_module('A', repo_id, **{'depend_ids': module_A_dependencies})
        self.assertTrue(module_A.is_iap, msg="depends on iap module")

    def test_commercial_packaged_version(self):
        """
        This test ensure that the commercial version are correctly computed and
        that pakage version return the version associated to a commerical version.
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_12 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_12').id)
        repo_13 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_13').id)

        module_11 = self._create_module('A', repo_11)
        self.assertEqual(module_11.id, module_11._get_commercial_series_module_id().id)

        template = module_11.template_id
        self.assertEqual(module_11.id, template.get_packaged_version(module_11).id)

        module_12 = self._create_module('A', repo_12)
        self.assertEqual(module_12.id, module_11._get_commercial_series_module_id().id)
        self.assertEqual(module_12.id, module_12._get_commercial_series_module_id().id)

        package_12 = template.get_packaged_version(module_12).ids
        self.assertEqual(set(package_12), set([module_11.id, module_12.id]))

        module_13 = self._create_module('A', repo_13)
        self.assertEqual(module_13.id, module_13._get_commercial_series_module_id().id)

        package_13 = template.get_packaged_version(module_13).id
        self.assertEqual(module_13.id, package_13)

    def test_assign_template(self):
        """
        This test ensure that two module sharing the same name and developed by the
        same maintainer will be associated to the same template.
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_12 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_12').id)
        repo_13 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_13').id)

        module_11 = self._create_module('A', repo_11)
        template = module_11.template_id

        template_12_A = self.env['loempia.module']._assign_template('A', repo_12)
        self.assertEqual(template_12_A.id, template.id, msg='Same name and Same maintainer refers to a same template')
        template_12_B = self.env['loempia.module']._assign_template('B', repo_12)
        self.assertTrue(template_12_B.id != template.id, msg='Not the same name')

        repo_13.write({'maintainer_id': self.manager.id})  # change the maintainer to create another template
        template_13_A = self.env['loempia.module']._assign_template('A', repo_13)
        self.assertTrue(template_13_A.id != template.id, msg='other maintainer means other template')

    def test_assign_product(self):
        """
        Test product assignation for older versions of a same module
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_12 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_12').id)
        repo_13 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_13').id)

        module_11 = self._create_module('A', repo_11)
        template = module_11.template_id
        module_12 = self._create_module('A', repo_12)
        module_13 = self._create_module('A', repo_13)

        module_11.product_id = 1
        module_13.product_id = 3  # Since v 13.0, each module is sold separately

        template._update_previous_version(module_11, update_product=True, update_pricing=False, update_purchases=False)
        template._update_previous_version(module_13, update_product=True, update_pricing=False, update_purchases=False)

        self.assertEqual(module_11.product_id, 1, msg='Module 11 and module 12 share the same product_id')
        self.assertEqual(module_12.product_id, 1, msg='Module 11 and module 12 share the same product_id')

        module_12.product_id = 2
        template._update_previous_version(module_12, update_product=True, update_pricing=False, update_purchases=False)

        self.assertEqual(module_11.product_id, 2, msg='Module 12 should keep the same product_id')
        self.assertEqual(module_12.product_id, 2, msg='Module 12 should keep the same product_id')

        self.assertEqual(module_13.product_id, 3, msg='Module 13 has a singular product_id')

    def test_update_pricing(self):
        """
        Test pricing update for older versions of a same module
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_12 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_12').id)
        repo_13 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_13').id)

        module_11 = self._create_module('A', repo_11)
        template = module_11.template_id
        module_12 = self._create_module('A', repo_12)
        module_13 = self._create_module('A', repo_13)

        module_11.price_single = 20.0
        module_11.currency_id = self.env.ref('base.USD').id

        template._update_previous_version(module_12, update_product=False, update_pricing=True, update_purchases=False)
        self.assertEqual(module_11.price_single, module_12.price_single, msg='Same product')
        self.assertEqual(module_11.currency_id.id, module_12.currency_id.id, msg='Same product')

        module_13.price_single = 12.0  # Since v 13.0, each module is sold separately
        template._update_previous_version(module_13, update_product=False, update_pricing=True, update_purchases=False)
        self.assertEqual(module_13.price_single, 12.0)
        self.assertEqual(module_12.price_single, 10.0)
        self.assertEqual(module_11.price_single, 10.0)

    def test_update_old_purchase(self):
        """
        Test purchase assignation for older versions of a same module
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_12 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_12').id)

        module_11 = self._create_module('A', repo_11)
        template = module_11.template_id
        module_12 = self._create_module('A', repo_12)

        po = self._create_confirmed_purchase(module_11)

        template._update_previous_version(module_12, update_product=False, update_pricing=False, update_purchases=True)

        self.assertEqual(po.module_id.id, module_12.id)

    def test_namespace_integrity(self):
        """
        In this test, we ensure that another maintainer cannot create a module if a module with the same name and
        same version already exists.
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_11_fake = self.env['loempia.repo'].create({
            'series_id': self.env.ref('loempia.series_11').id,
            'url': 'ssh://git@github.com/ndeodoo/essai2.git#11.0',
            'vcs': 'git',
            'maintainer_id': self.manager.id,
            'state': 'active',
        })

        module_11 = self._create_module('A', repo_11)
        is_available = module_11._is_name_available_for_serie(repo_11_fake)
        self.assertTrue(not is_available)
        is_available = module_11._is_name_available_for_serie(repo_11)
        self.assertTrue(is_available)

    def test_techname_integrity(self):
        """
        In this test, we ensure that a techname is correctly reserved
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        repo_12 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_12').id)
        repo_13 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_13').id)

        module_11 = self._create_module('A', repo_11)
        module_13 = self._create_module('A', repo_13)

        modules = module_11 + module_13
        is_techname_available = modules._is_techname_available(repo_12, {'name': 'A'})
        self.assertTrue(is_techname_available)

        repo_12.maintainer_id = self.manager.id
        repo_12.url = 'ssh://git@github.com/ndeodoo/essai2.git#11.0'

        is_techname_available = modules._is_techname_available(repo_12, {'name': 'A'})
        self.assertTrue(not is_techname_available)

    def test_validate_license(self):
        """
        In this test, we ensure that an erroneous license can't be set.
        """
        repo_11 = self.repositories.filtered(lambda r: r.series_id.id == self.env.ref('loempia.series_11').id)
        Module = self.env['loempia.module']

        # 01 Fake license
        with self.assertRaises(ManifestError):
            values = {'license': 'ldp-e'}  # Fake license
            Module._validate_license(repo_11, values)

        # 02 OEEL can not be used for third party apps
        with self.assertRaises(ManifestError):
            values = {'license': 'OEEL-1'}  # OEEL-1 is not allowed as license
            Module._validate_license(repo_11, values)

        # 02 Server repo
        repo_11.auto_approve = True
        repo_11.url = "ssh://git@github.com/odoo/odoo.git#11.0"
        values = {'license': 'Other proprietary'}
        Module._validate_license(repo_11, values)

        self.assertEqual(values['license'], 'LGPL-3', msg="Server repo after v11 should have LGPL-3 license")

        # 03 Enterprise repo
        repo_11.url = "ssh://git@github.com/odoo/enterprise.git#11.0"
        values = {'license': 'Other proprietary'}
        Module._validate_license(repo_11, values)

        self.assertEqual(values['license'], 'OEEL-1', msg='Enterprise repo')

    def test_module_recursion(self):
        """Test that we can't have cyclic dependencies between modules."""
        repo = self.repositories[0]
        mod1 = self._create_module('mod1', repo)
        mod2 = self._create_module('mod2', repo)
        mod1.write({'depend_ids': [Command.link(mod2.id)]})
        with self.assertRaises(ValidationError):
            mod2.write({'depend_ids': [Command.link(mod1.id)]})

    def test_module_unlink_with_interdependency(self):
        """Test that we can't delete a module if other modules still depends on it."""
        repo = self.repositories[0]
        mod2 = self._create_module('mod2', repo)
        mod1 = self._create_module('mod1', repo, depend_ids=[
            Command.set([mod2.id]),
        ])

        # set dummy product id
        mod1.product_id = -1
        mod2.product_id = -2

        # removing only mod2 should be raise an error
        with patch.object(type(mod2), '_disable_module_product',
                          autospec=True, return_value=None) as disable_module_product:
            with self.assertRaises(psycopg2.errors.ForeignKeyViolation), mute_logger('odoo.sql_db'):
                mod2.unlink()
            # in case of failure, we should not call accounts. to disable the
            # module's product
            disable_module_product.assert_not_called()

        # but removing both mod1 and mod2 should works
        with patch.object(type(mod2), '_disable_module_product',
                          autospec=True, return_value=None) as disable_module_product:
            (mod1 | mod2).unlink()
            # check that we've called prod to disable the product there
            disable_module_product.assert_called()
            self.assertEqual(disable_module_product.call_count, 2)

    def test_module_unlink_batch(self):
        """Test that deleting a modules (and all modules that depends on it) works."""
        repo = self.repositories[0]
        mod1 = self._create_module('mod1', repo)
        mod2 = self._create_module('mod2', repo)
        mod3 = self._create_module('mod3', repo, depend_ids=[
            Command.set([mod1.id, mod2.id]),
        ])

        # deleting a module and all modules depending on it should not raise an error
        (mod1 | mod3).unlink()
        self.assertEqual((mod1 | mod2 | mod3).exists(), mod2)

        # deleting a single module with no upstream dependencies
        mod2.unlink()
        self.assertFalse(mod2.exists())

    def test_cloc_industry_module(self):
        """Test that loempia cloc and standard cloc give the same result."""
        dummy_path = os.path.dirname(__file__) + '/'
        dummy_module = 'dummy_cloc'
        zip_module = BytesIO()
        with ZipFile(zip_module, "w") as zf:
            for dirname, _subdirs, files in os.walk(dummy_path + dummy_module):
                zip_dirname = os.sep.join(dirname.split(dummy_path)[1:])
                zf.write(dirname, zip_dirname)
                for filename in files:
                    zf.write(os.path.join(dirname, filename), os.path.join(zip_dirname, filename))
        self.env['ir.module.module'].sudo()._import_zipfile(zip_module)
        loempia_cloc = util._count_lines(dummy_path + dummy_module)
        self.assertNotEqual(loempia_cloc, 0, "Maintenance of dummy module should not be free of charge")
        # force re-reading the views into the cache to remove the xml headers
        self.env['ir.ui.view'].invalidate_model()
        cl = cloc.Cloc()
        cl.count_customization(self.env)
        self.assertEqual(
            loempia_cloc, cl.code.get(dummy_module, 0), "Standard cloc and loempia cloc codes diverge",
        )
